// fl_autorun, Copyright 2007, Don Zimmer, dzimmer@magma.ca
// public domain 
//
#pragma warning ( disable : 4996 4312 4311 )
#pragma comment ( lib , "shell32.lib")
#pragma comment ( lib , "user32.lib" )
#pragma comment ( lib , "winmm.lib"  )
#pragma comment ( lib , "user32.lib" )
#pragma comment ( lib , "gdi32.lib"  )
#include <windows.h>
#include <mmsystem.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <direct.h>
#include <ctype.h>

#include <FL/Fl.H>
#include <FL/x.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Menu_Item.H>
#include <FL/fl_ask.H>
#include <FL/Fl_Shared_Image.H>

#include "Fl_DragBar.H"
#include "Fl_Multi_Widget.H"
#include "resource.h"
#include "parse.h"

static uchar        def_labelsize       = 14;
static Fl_Font      def_labelfont       = FL_HELVETICA;
static Fl_Labeltype def_labeltype       = FL_NORMAL_LABEL;
static Fl_Color     def_labelcolor      = FL_BLACK;
static Fl_Color     def_color           = FL_BACKGROUND_COLOR;
static Fl_Color     def_selection_color = FL_SELECTION_COLOR;
static Fl_Boxtype   def_box             = FL_UP_BOX;
static Fl_Boxtype   def_down_box        = FL_NO_BOX;
static Fl_Align     def_align           = FL_ALIGN_CENTER;

static uchar        labelsize       = def_labelsize;
static Fl_Font      labelfont       = def_labelfont;
static Fl_Labeltype labeltype       = def_labeltype;
static Fl_Color     labelcolor      = def_labelcolor;
static Fl_Color     color           = def_color;
static Fl_Color     selection_color = def_selection_color;
static Fl_Boxtype   box             = def_box;
static Fl_Boxtype   down_box        = def_down_box;
static Fl_Align     align           = def_align;

static void set_attributes(Fl_Multi_Widget *o)
{
  o->box(box);
  o->down_box(down_box);
  o->color(color);
  o->selection_color(selection_color);
  o->align(align);
  o->labelcolor(labelcolor);
  o->labelsize(labelsize);
  o->labelfont(labelfont);

  if(!o->image()) o->labeltype(labeltype);

  // added so scheme background can show through
  // 
  Fl_Window *w = o->window();
  if(w && (o == w->child(0)))
  {
    w->box(o->box());
    w->color(o->color());
    o->box((FL_FLAT_BOX == box)?FL_NO_BOX:fl_frame(box)); 
  }
}

static int syntax_error(const char *key, const char *arg, const char *hint)
{
  if(!key) key = "";
  if(!arg) arg = "";
  if(!hint) hint = "";
  fl_alert("Syntax Error: %s\nArguments: %s\nHint: %s", key, arg, hint);
  return -1;
}

static void init(void)
{
  fl_define_FL_ROUND_UP_BOX();
  fl_define_FL_SHADOW_BOX();
  fl_define_FL_ROUNDED_BOX();
  fl_define_FL_RFLAT_BOX();
  fl_define_FL_RSHADOW_BOX();
  fl_define_FL_DIAMOND_BOX();
  fl_define_FL_OVAL_BOX();
  fl_define_FL_PLASTIC_UP_BOX();

  fl_define_FL_SHADOW_LABEL();
  fl_define_FL_ENGRAVED_LABEL();
  fl_define_FL_EMBOSSED_LABEL();

  fl_register_images();

  char path[BUFSIZE];
  GetModuleFileName(0, path, BUFSIZE);
  char *slash = strrchr(path, '\\');
  slash[1] = 0;
  chdir(path);
}

static void cb_minimize(Fl_Widget *o, void *v)
{
  o->window()->iconize();
}

static void cb_hide(Fl_Widget *o, void *v)
{
  o->window()->hide();
}

static void cb_winclose(Fl_Widget *o, void *v)
{
  if(Fl::event()!=FL_SHORTCUT || Fl::event_key()!=FL_Escape) 
  {
    exit(0);
  }
}

static void cb_widget(Fl_Widget *o, void *v)
{
  char *cmd = (char*)v;

  if(cmd[0])
  {
    char *dot = strrchr(cmd, '.');

    if(!strcmp(cmd, "FL_EXIT"))
    {
      o->window()->hide();
    }
    else if(dot && !stricmp(dot, ".wav"))
    {
      Fl_Multi_Widget *m = (Fl_Multi_Widget*)o;
      if(m->value()) 
      {
        PlaySound(cmd, 0, SND_FILENAME | SND_ASYNC | SND_LOOP); 
      }
      else 
      {
        PlaySound(0, 0, SND_PURGE); 
      }
    }
    else
    {
      char *comma = strchr(cmd, ',');
      if(comma) *comma = '\0';
      ShellExecute(0, "open", cmd, (comma)?comma+1:0, 0, SW_SHOW);
      if(comma) *comma = ',';
    }
  }
}

static void add_dragbar(Fl_Boxtype box)
{
  Fl_DragBar *bar = new Fl_DragBar(0, 0, 0, 0);

  Fl_Window *window = bar->window();

  // user did not specify a dragbar, so make background draggable
  //
  if(-1 == box)
  {
    bar->size(window->w(), window->h());
    bar->box(FL_NO_BOX);
    return;
  }

  Fl_Widget *form = window->child(0);

  // only left or center
  Fl_Align bar_align = (Fl_Align)(form->align() & (FL_ALIGN_LEFT | FL_ALIGN_CENTER)); 

  // limit font size to fit
  uchar bar_labelsize = (form->labelsize() > 16)? 16: form->labelsize();

  bar->size(window->w(), 25);
  bar->label(window->label());
  bar->box(box);
  bar->color(form->color());
  bar->align(bar_align);
  bar->labeltype(form->labeltype());
  bar->labelcolor(form->labelcolor());
  bar->labelsize(bar_labelsize);
  bar->labelfont(form->labelfont());
 
  Fl_Button *btn = new Fl_Button(bar->w()-50, 0, 25, 23);
  btn->box(FL_NO_BOX);
  btn->label("@#+42>");
  btn->labelcolor(bar->color());
  btn->callback(cb_minimize);

  btn = new Fl_Button(bar->w()-25, 0, 25, 25);
  btn->box(FL_NO_BOX);
  btn->label("@#3+");
  btn->labelcolor(bar->color());
  btn->callback(cb_hide);
}

// make borderless toplevel window show in taskbar with window menu
//
static void fix_borderless(Fl_Window *w)
{
#ifdef _WIN32
#pragma warning ( push )
#pragma warning ( disable : 4244 )
#if (_MSC_VER < 1300)
#define LONG_PTR LONG
# define GetWindowLongPtr GetWindowLong
# define SetWindowLongPtr SetWindowLong
#endif

  if(w->border()) return;

  HWND wnd = fl_xid(w);
  if(!wnd)
  {
    Fl_Group *grp = Fl_Group::current();
    Fl_X::make(w);
    Fl_Group::current(grp);
    wnd = fl_xid(w);
  }

  LONG_PTR styleEx = GetWindowLongPtr(wnd, GWL_EXSTYLE);
  styleEx &= ~WS_EX_TOOLWINDOW;
  SetWindowLongPtr(wnd, GWL_EXSTYLE, styleEx);

  LONG_PTR style = GetWindowLongPtr(wnd, GWL_STYLE);
  style |= WS_SYSMENU;
  SetWindowLongPtr(wnd, GWL_STYLE, style);

#pragma warning ( pop )
#endif
}

// special case to allow non-rectangular windows
//
static void clip_borderless(Fl_Window *w)
{
  if(w->border()) return;

  HWND wnd = fl_xid(w);
  if(!wnd)
  {
    Fl_Group *grp = Fl_Group::current();
    Fl_X::make(w);
    Fl_Group::current(grp);
    wnd = fl_xid(w);
  }

  Fl_Boxtype box = w->box();

  if(FL_RFLAT_BOX == box)
  {
    int minDim = (w->w() < w->h())? w->w() : w->h();
    HRGN hRgn = CreateRoundRectRgn(0, 0, w->w(), w->h(), minDim, minDim);
    SetWindowRgn(wnd, hRgn, TRUE);
    w->box(FL_FLAT_BOX);
  }
  else if(FL_OFLAT_BOX == box)
  {
    HRGN hRgn = CreateEllipticRgn(0, 0, w->w(), w->h());
    SetWindowRgn(wnd, hRgn, TRUE);
    w->box(FL_FLAT_BOX);
  }
}

int CALLBACK WinMain(HINSTANCE hInstance,
    HINSTANCE hPrevInstance,
    LPSTR lpCmdLine,
    int nCmdShow)
{
  int x, y, w, h, idx, audio = 0;
  char key[BUFSIZE], arg[BUFSIZE], label[BUFSIZE];
  FILE *fp;
  Fl_Multi_Widget *o, *push = 0;
  Fl_Window *window = 0;
  Fl_Boxtype dragbox = (Fl_Boxtype)-1;

  init();

  fp = fopen("autorun.ini", "rt");
  if(!fp) 
  {
    fl_alert("Missing \"autorun.ini\"");
    return -1;
  }

  if(!readln(fp, key, arg) || stricmp(key, "fl_autorun"))
  {
    return syntax_error(key, arg, "expecting \"fl_autorun\"");
  }

  do
  {
    if(!stricmp(key, "fl_autorun"))
    {
      if(window)
      {
        return syntax_error(key, arg, "keyword duplicated");
      }
      if(!parse_widget(arg, &x, &y, &w, &h, label)) 
      {
        return syntax_error(key, arg, "x y w h [l]");
      }
      window = new Fl_Double_Window(x, y, w, h, (label[0])?strdup(label):0);
      window->icon((char*)LoadIcon(fl_display, MAKEINTRESOURCE(IDI_CDROM)));
      o = new Fl_Multi_Widget(0, 0, w, h);
      idx = 0;
    }
    else if(!stricmp(key, "widget"))
    {
      set_attributes(o);
      if(!window->border() && window->child(0) == o) add_dragbar(dragbox);
      idx++;
      if(!parse_widget(arg, &x, &y, &w, &h, label)) 
      {
        return syntax_error(key, arg, "x y w h [l]");
      }
      o = new Fl_Multi_Widget(x, y, w, h, (label[0])?strdup(label):0);
    }
    else if(!stricmp(key, "background"))
    {
      uchar r, g, b;
      Fl::get_color(parse_color(arg, FL_BACKGROUND_COLOR), r, g, b);
      Fl::background(r, g, b);
    }
    else if(!stricmp(key, "background2"))
    {
      uchar r, g, b;
      Fl::get_color(parse_color(arg, FL_BACKGROUND2_COLOR), r, g, b);
      Fl::background2(r, g, b);
    }
    else if(!stricmp(key, "foreground"))
    {
      uchar r, g, b;
      Fl::get_color(parse_color(arg, FL_FOREGROUND_COLOR), r, g, b);
      Fl::foreground(r, g, b);
    }
    else if(!stricmp(key, "syscolors"))
    {
      Fl::get_system_colors();
    }
    else if(!stricmp(key, "scheme"))
    {
      const char *scheme = parse_scheme(arg, "none");
      Fl::scheme(scheme);
    }
    else if(!stricmp(key, "noborder"))
    {
      window->border(0);
    }
    else if(!stricmp(key, "icon"))
    {
      HANDLE hIcon = LoadImage(0, arg, IMAGE_ICON, 0, 0, LR_LOADFROMFILE);
      if(hIcon) window->icon(hIcon);
    }
    else if(!stricmp(key, "audio"))
    {
      audio = 1;
    }
    else if(!stricmp(key, "noescape"))
    {
      window->callback(cb_winclose);
    }
    else if(!stricmp(key, "nofocus"))
    {
      Fl::visible_focus(0);
    }
    else if(!stricmp(key, "dragbar"))
    {
      dragbox = parse_box(arg, FL_NO_BOX);
    }
    else if(!stricmp(key, "box"))
    {
      box = parse_box(arg, def_box);
    }
    else if(!stricmp(key, "down_box"))
    {
      down_box = parse_box(arg, def_down_box);
    }
    else if(!stricmp(key, "color"))
    {
      color = parse_color(arg, def_color);
    }
    else if(!stricmp(key, "selection_color"))
    {
      selection_color = parse_color(arg, def_selection_color);
    }
    else if(!stricmp(key, "align"))
    {
      align = parse_align(arg, def_align);
    }
    else if(!stricmp(key, "labeltype"))
    {
      labeltype = parse_labeltype(arg, def_labeltype);
    }
    else if(!stricmp(key, "labelcolor"))
    {
      labelcolor = parse_color(arg, def_labelcolor);
    }
    else if(!stricmp(key, "labelsize"))
    {
      labelsize = parse_labelsize(arg, def_labelsize);
    }
    else if(!stricmp(key, "labelfont"))
    {
      labelfont = parse_labelfont(arg, def_labelfont);
    }
    else if(!stricmp(key, "picture"))
    {
      o->picture(Fl_Shared_Image::get(arg));
    }
    else if(!stricmp(key, "texture"))
    {
      o->texture(Fl_Shared_Image::get(arg));
    }
    else if(!stricmp(key, "image"))
    {
      Fl_Labeltype ll = o->labeltype();
      o->image(Fl_Shared_Image::get(arg));
      Fl_Labeltype mm = o->labeltype();
      int a = 6;
    }
    else if(!stricmp(key, "tooltip"))
    {
      o->tooltip(strdup(arg));
    }
    else if(!stricmp(key, "callback"))
    {
      o->callback(cb_widget, strdup(arg));

      char *dot = strrchr(arg, '.');

      if(dot && !stricmp(dot, ".wav"))
      {
        o->type(FL_TOGGLE_BUTTON);
        push = o;
      }
    }
    else if(key[0] && !ispunct(key[0]))
    {
      return syntax_error(key, arg, "unsupported keyword");
    }
  }
  while(readln(fp, key, arg));

  fclose(fp);

  set_attributes(o);
  fix_borderless(window);
  clip_borderless(window);
  if(push && audio) { push->set(); push->do_callback(); }
  window->show();
  return Fl::run();
}

